;****************************************************************************
;*                                                                          *
;* StarEMU  ver.2.50                                                        *
;*                                                                          *
;* StarFlight I     VGA 160x200, 16-color active emulator                   *
;*                                                                          *
;****************************************************************************
; +------------------------------------------------------------------------+
; |                                                                        |
; | EmuCode.ASM    Version 2.25    (C) Jouko Orava, 1996                   |
; |                                                                        |
; +------------------------------------------------------------------------+
;
; Compile to .COM format. Use EXE2BIN, if compiled to .EXE format.
;
.286
.model tiny
.code
 org 100h          ; This is .COM Code
;============================================================================
         Origin:
               jmp   DosEntry
               db    01Ah              ; Should waste total four bytes here.
;============================================================================

               Ticks dw 00000h
            TickSize dw 00008h         ; rate = ticksize*1024/ticklimit
           TickLimit dw 00075h         ; curr = 70.01 Hz

              Frames dw 00000h         ; RTC is updated every FrameRefresh
          FrameLimit dw 07FFFh         ; frames

           OldInt70h dw 00000h,00000h  ; Int 70h vector (stored)
           OldInt15h dw 00000h,00000h  ; Int 15h vector (stored)
           OldInt10h dw 00000h,00000h  ; Int 10h vector (stored)

               Paras dw 00000h         ; Paragraphs reserved when TSR'ed
            ModeFlag dw 00000h         ; Lowest bit 1 if emulation enabled

        COMStackSize EQU 0200h         ; Temporary stack size.

               OnKey EQU 058h          ; F12   (lshift=2A,F2=3C)
              OffKey EQU 057h          ; F11   (ctrl=1D,F1=3B)
             JunkKey EQU 01Dh          ; CTRL  (alt=38,F3=3D)

;+-----+ +--------------------------+ +-------------------------------------+
;|     | | Mode Initialization Code | |                                     |
;+-----+ +--------------------------+ +-------------------------------------+

       InitMode:
               push  ax                ; Save registers AX,BX,DX,SI & DS
               push  bx
               push  dx
               push  si
               push  ds

               mov   ax,00083h         ; Set text mode 3, no memory clr
               int  0A1h               ; Old int 10h
               mov   ax,01000h         ; Set Horizontal Panning to zero
               mov   bx,00013h
               int  0A1h               ; Old int 10h
               mov   ax,01003h         ; Set No Blinking
               xor   bx,bx
               int  0A1h               ; Old int 10h

               mov   dx,003C2h         ; Set 25MHz Clock
               mov   al,063h
               out   dx,al

               mov   dx,003D4h
               mov   ax,0C009h         ; Set Double scans + font height to 1
               out   dx,ax

               mov   dx,003C4h
               mov   ax,00101h         ; Set font width to 8
               out   dx,ax
               mov   ax,00503h         ; Set font bank 1
               out   dx,ax
               mov   ax,00C02h         ; Set write plane mask (2&3 ok)
               out   dx,ax

               mov   dx,003CEh
               mov   ax,00004h         ; Set read plane to 0
               out   dx,ax

               mov   ax,0B800h         ; Video segment is B800
               mov   ds,ax
               xor   si,si             ; Prepare to copy 4000h bytes

          Copy1:
               mov   ax,[si+00h]       ; Copy to itself; same source and
               mov   bx,[si+02h]       ;  target. Copies data across
               mov   [si+00h],ax       ;  planes.
               mov   [si+02h],bx
               mov   ax,[si+04h]
               mov   bx,[si+06h]
               mov   [si+04h],ax
               mov   [si+06h],bx
               add   si,08h
               cmp   si,04000h
               jb   Copy1

               mov   ax,00F0Fh         ; Low bits define right-side pixel

          Fill1:
               mov   [si],ax           ; Fill with font data. Some
               mov   [si+02h],ax       ;  incompatible videocards may not
               mov   [si+04h],ax       ;  work with this direct approach;
               mov   [si+06h],ax       ;  a 256-byte buffer (filled 0Fh)
               mov   [si+08h],ax       ;  and int10h/character generator
               mov   [si+0Ah],ax       ;  call to replace all 256 fonts
               mov   [si+0Ch],ax       ;  with one-byte fonts has to be
               mov   [si+0Eh],ax       ;  used.
               add   si,010h           ;
               jns  Fill1

               mov   dx,003CEh
               mov   ax,00204h         ; Set Read Plane to 2
               out   dx,ax

               pop   ds                ; Restore registers
               pop   si
               pop   dx
               pop   bx
               pop   ax

               ret

;+-----+ +-------------------+ +--------------------------------------------+
;|     | | Mode Closeup Code | |                                            |
;+-----+ +-------------------+ +--------------------------------------------+

        CloseUp:
               push  ax                ; Save registers
               push  dx
               push  si
               push  ds

               mov   ax,0B800h         ; Video segment is 0B800h
               mov   ds,ax

               mov   dx,003C4h
               mov   ax,00F02h         ; Set write plane mask (write to all)
               out   dx,ax

               xor   si,si             ; Prepare to copy 4000h bytes
          Copy2:
               mov   ax,[si]           ; Read from plane 2 and write the
               mov   dx,[si+02h]       ;  data to all planes.
               mov   [si],ax
               mov   [si+02h],dx
               mov   ax,[si+04h]
               mov   dx,[si+06h]
               mov   [si+04h],ax
               mov   [si+06h],dx
               add   si,08h
               cmp   si,04000h
               jb   Copy2

               mov   ax,00085h         ; Restore mode 5 (320x200 CGA 4-color)
               int  0A1h               ; Old Int 10h

               pop   ds                ; Restore registers
               pop   si
               pop   dx
               pop   ax

               ret

;+-----+ +-------------------+ +--------------------------------------------+
;|     | | Video Update Code | |                                            |
;+-----+ +-------------------+ +--------------------------------------------+

    VideoUpdate:
               push  ax                ; Save registers
               push  bx
               push  cx
               push  dx
               push  si
               push  di
               push  ds

               mov   ax,0B800h         ; Video segment is B800h
               mov   ds,ax

               mov   dx,003C4h
               mov   ax,00302h         ; Write to planes 0&1 (visible)
               out   dx,ax

               xor   si,si             ; = Source offset
               xor   di,di             ; = Target offset
               mov   al,014h           ; = pixel groups at a scan line
               mov   cl,al             ; (initialize pixel group count)

         VULoop:
               mov   bx,[si]           ; Load source (interleave also),
               mov   ah,bl             ;  note that target has valid data
               mov   dx,[si+02h]       ;  in odd bytes. Even targets are
               mov   ch,dl             ;  ignored.
               mov   [di],ax           ; Store interleaved source data..
               mov   [di+02h],bx       ;  :
               mov   [di+04h],cx       ;  :
               mov   [di+06h],dx       ; Ok, target stored.
               add   si,02000h         ; Adjust for odd scan (8k bank)
               add   di,0A0h           ; Target only A0h bytes, of course
               mov   bx,[si]           ; Repeat the copy process for this
               mov   ah,bl             ;  odd-scan piece also.
               mov   dx,[si+02h]       ;  :
               mov   ch,dl             ;  :
               mov   [di],ax           ; (Store interleaved source data)
               mov   [di+02h],bx       ;  :
               mov   [di+04h],cx       ;  :
               mov   [di+06h],dx       ; Done.
               sub   si,01FFCh         ; Back to even scan, next pix group.
               sub   di,098h           ;  (Target, too.)
               dec   cl                ; Is last group on scan?
               jnz  VULoop
               add   di,0A0h           ; Next scan for target (source is ok)
               mov   cl,al             ; Reset pixel group count
               cmp   si,01F40h         ; Is all done yet?
               jb   VULoop

               mov   dx,003C4h
               mov   ax,00C02h         ; Write to planes 2&3 (invisible)
               out   dx,ax

               pop   ds                ; Restore registers
               pop   di
               pop   si
               pop   dx
               pop   cx
               pop   bx
               pop   ax

               ret

;+-----+ +---------------------+ +------------------------------------------+
;|     | | Video Clear routine | |                                          |
;+-----+ +---------------------+ +------------------------------------------+

       VideoCLR:
               push  ax                ; Save registers
               push  si
               push  ds

               mov   ax,0B800h         ; Video segment is B800h
               mov   ds,ax
               xor   ah,ah
               xor   si,si

        CLRLoop:
               mov   [si],ax
               mov   [si+002h],ax
               mov   [si+004h],ax
               mov   [si+006h],ax
               mov   [si+008h],ax
               mov   [si+00Ah],ax
               mov   [si+00Ch],ax
               mov   [si+00Eh],ax
               add   si,010h
               cmp   si,03F40h
               jb    CLRLoop

               pop   ds                ; Restore registers
               pop   si
               pop   ax

               ret
;+-----+ +--------------------------+ +-------------------------------------+
;|     | | ID Just Before NewInt10h | |                                     |
;+-----+ +--------------------------+ +-------------------------------------+
     Align Word
               id_ST EQU 05374h
               id_AR EQU 06172h
               id_EM EQU 0454Dh
               id_U2 EQU 05532h
               id_50 EQU 03530h
               dw   id_ST,id_AR,id_EM,id_U2,id_50
;+-----+ +----------------------------+ +-----------------------------------+
;|     | | New Int10h Service Routine | |                                   |
;+-----+ +----------------------------+ +-----------------------------------+
      NewInt10h:
               cmp   ah,000h           ; Is *Mode Set* ?
               jne  N10_Out

               cmp   al,005h           ; Should I enable ME?
               je   N10_On
               cmp   al,004h           ; Should I enable ME?
               je   N10_On

               test  byte ptr CS:[ModeFlag],001h
               jnz  N10_Off

        N10_Out:
               jmp   dword ptr CS:[OldInt10h]

         N10_On:
               mov   al,005h
               int  0A1h               ; Old Int 10h
               call InitMode
               call VideoCLR
               call VideoUpdate
               or    byte ptr CS:[ModeFlag],001h
               iret

        N10_Off:
               and   byte ptr CS:[ModeFlag],0FEh
               call CloseUp
               jmp   dword ptr CS:[OldInt10h]

;+-----+ +----------------------------+ +-----------------------------------+
;|     | | New Int15h Service Routine | |                                   |
;+-----+ +----------------------------+ +-----------------------------------+
      NewInt15h:
               pushf
               cmp   ax,04F00h+OnKey
               je   N15_On
               cmp   ax,04F00h+OffKey
               je   N15_Off
               cmp   ax,04F00h+JunkKey
               je   N15_Junk

        N15_Out:
               popf
               jmp   dword ptr CS:[OldInt15h]

         N15_On:
               test  byte ptr CS:[ModeFlag],001h
               jnz  N15_Update
               call InitMode
               or    byte ptr CS:[ModeFlag],001h

     N15_Update:
               call VideoUpdate
               popf
               clc                             ; Lose the keypress
               jmp   dword ptr CS:[OldInt15h]

        N15_Off:
               test  byte ptr CS:[ModeFlag],001h
               jz   N15_Out

               and   byte ptr CS:[ModeFlag],0FEh
               call CloseUp
               popf
               clc                             ; Lose the keypress
               jmp   dword ptr CS:[OldInt15h]

       N15_Junk:
               test  byte ptr CS:[ModeFlag],001h
               jz   N15_Out
               call CloseUp
               call InitMode
               call VideoUpdate
               popf
               clc                             ; Lose the keypress
               jmp   dword ptr CS:[OldInt15h]

;+-----+ +----------------------------+ +-----------------------------------+
;|     | | New Int70h Service Routine | |                                   |
;+-----+ +----------------------------+ +-----------------------------------+

      NewInt70h:
               int  0A0h
               push  ax

               mov   ax,word ptr CS:[Ticks]     ; Get Ticker
               add   ax,word ptr CS:[TickSize]  ; calc next tick .. eh!
               cmp   ax,word ptr CS:[TickLimit] ; is tick overflow?
               jge  N70_Go

               mov   word ptr CS:[Ticks],ax     ; no, update ticker
               pop   ax
               iret                             ; and done.

         N70_Go:
               sub   ax,word ptr CS:[TickLimit] ; Update Ticker
               mov   word ptr CS:[Ticks],ax

               inc   word ptr CS:[Frames]       ; increase frame counter
               mov   ax,word ptr CS:[FrameLimit]
               cmp   word ptr CS:[Frames],ax
               jb   N70_Update

               xor   ax,ax
               push  bx
               push  cx
               push  dx
               push  es
               mov   word ptr CS:[Frames],ax
               mov   ax,08301h                  ; disable interval
               int  015h
               mov   ax,08300h
               mov   bx,cs
               mov   es,bx
               mov   bx,offset ModeFlag+1
               mov   cx,07FFFh
               mov   dx,cx
               int  015h
               pop   es
               pop   dx
               pop   cx
               pop   bx

     N70_Update:
               pop   ax
               test  byte ptr CS:[ModeFlag],001h
               jz   N70_Outta
               call VideoUpdate
      N70_Outta:
               iret



;+-----+ +---------------------+ +------------------------------------------+
;|     | | Main and other code | |                                          |
;+-----+ +---------------------+ +------------------------------------------+
    NotResident:
    Msg01 db 'StarEMU   v2.50   (C) Jouko Orava, 1996'
          db 0Ah,0Dh,24h

    Msg02 db 'StarEMU is now resident and will activate on CGA modes',0Ah,0Dh
          db ' 4 and 5 (320x200, 4 colors), and deactivate for other',0Ah,0Dh
          db 'video modes. There are manual overrides:',0Ah,0Dh
          db ' F11 : Deactivate mode (data will NOT be lost)',0Ah,0Dh
          db '       (Back to normal CGA 320x200 4-color ugh)',0Ah,0Dh
          db ' F12 : Activate mode (use only when in CGA video',0Ah,0Dh
          db '        modes, as deactivation always returns to',0Ah,0Dh
          db '        320x200,4-color CGA mode.)',0Ah,0Dh
          db 'CTRL : Junk-discarder. If StarFlight messes up with',0Ah,0Dh
          db '       your display, press CTRL; this should fix it.',0Ah,0Dh
          db 024h

    Msg03 db 'StarEMU is already resident. If you forgot the keys, they are:'
          db 0Ah,0Dh
          db ' F11 Deactivates, F12 Activates,  and Control (CTRL) discards'
          db ' junk.',0Ah,0Dh
          db 024h

       DosEntry:
               mov   ax,cs             ; Initialize all segment registers
               mov   ds,ax
               mov   bx,offset ComStack+ComStackSize
               cli                     ; Yes, I'm paranoid
               mov   ss,ax             ;  :
               mov   sp,bx             ;  :
               sti                     ; At least when using stacks

               mov   ax,word ptr CS:[002Ch]
               mov   es,ax             ; Free Environment segment
               mov   ax,04900h
               int  021h

               mov   ax,offset NotResident+00Fh    ; Calculate paras req'd
               shr   ax,4
               mov   word ptr CS:[offset Paras],ax

               mov   ah,009h           ; Show ID-message.
               mov   dx,offset MSG01
               int  021h

               mov   ax,03515h         ; Get Int15 vector
               int  021h
               mov   word ptr CS:[offset OldInt15h],bx
               mov   ax,es
               mov   word ptr CS:[offset OldInt15h+2],ax

               mov   ax,03510h         ; Get Int10 vector
               int  021h
               mov   word ptr CS:[offset OldInt10h],bx
               mov   ax,es
               mov   word ptr CS:[offset OldInt10h+2],ax

               dec   ax                ; Adjust segment ..
               mov   es,ax
               cmp   es:[bx+06h],id_ST ; Check ID
               jne  NotAlready
               cmp   es:[bx+08h],id_AR
               jne  NotAlready
               cmp   es:[bx+0Ah],id_EM
               jne  NotAlready
               cmp   es:[bx+0Ch],id_U2
               jne  NotAlready
               cmp   es:[bx+0Eh],id_50
               jne  NotAlready

               mov   ax,cs             ; Already resident
               mov   ds,ax
               mov   ax,0900h          ; message..
               mov   dx,offset Msg03
               int  021h

               mov   ax,04C00h         ; And out.
               int  021h

      NotAlready:
               mov   ax,word ptr CS:[offset OldInt10h+2]
               mov   ds,ax             ; intA1 = old int 10
               mov   ax,025A1h
               mov   dx,word ptr CS:[offset OldInt10h]
               int  021h
               mov   ax,cs
               mov   ds,ax

               mov   ax,03570h         ; Get Int70 vector
               int  021h
               mov   word ptr CS:[offset OldInt70h],bx
               mov   ax,es
               mov   word ptr CS:[offset OldInt70h+2],ax

               mov   ds,ax             ; intA0 = old int 70
               mov   ax,025A0h
               mov   dx,bx
               int  021h
               mov   ax,cs
               mov   ds,ax

               mov   ax,02515h         ; Set New Int15h handler
               mov   dx,offset NewInt15h
               int  021h

               mov   ax,02510h         ; Set New Int10h handler
               mov   dx,offset NewInt10h
               int  021h

               mov   ax,02570h         ; Set New Int70h handler
               mov   dx,offset NewInt70h
               int  021h

               mov   ax,08301h         ; Disable event wait
               int  015h
               mov   ax,08300h         ; Set maximum interval
               mov   bx,cs
               mov   es,bx
               mov   bx,offset ModeFlag+1
               mov   cx,07FFFh
               mov   dx,cx
               int  015h

               mov   ah,009h           ; Show all-ok -message.
               mov   dx,offset MSG02
               int  021h

               mov   ax,03100h         ; Terminate and stay resident
               mov   dx,word ptr DS:[offset Paras]
               int  021h

;+-----+ +----------------------+ +-----------------------------------------+
;|     | | Temporary .COM stack | |                                         |
;+-----+ +----------------------+ +-----------------------------------------+
     align word
       COMStack: db (COMStackSize) dup (00h)
 ends
 end Origin
